#define MODULE
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <asm/io.h>

#include <rtai.h>

const int SECS = 5;

MODULE_PARM(SECS, "i");

static int RESET_COUNT;

static void fun(void)
{
	static int count = 0, gcount = -1;
	static RTIME tbase;
	long linux_cr0, fpu_reg[27];
	double cpu_freq;
	union {unsigned long long time; unsigned long time_lh[2]; } tsc;
	tsc.time =  rd_CPU_ts();
	if (gcount < 0) {
		tbase = tsc.time;
	}
	gcount++;
	if (++count == RESET_COUNT) {
		tsc.time -= tbase;
		__asm__ __volatile__("movl %%cr0,%%eax": "=a" (linux_cr0): : "ax");
		__asm__ __volatile__("clts; fnsave %0": "=m" (fpu_reg));
		cpu_freq =  (double)tsc.time_lh[1]*(double)0x100000000LL + (double)tsc.time_lh[0];
		count =  (cpu_freq*CLOCK_TICK_RATE)/(((double)gcount)*LATCH) + 0.4999999999999;
		__asm__ __volatile__("frstor %0" : "=m" (fpu_reg));
		__asm__ __volatile__("movl %%eax,%%cr0": :"a" (linux_cr0): "ax");
		printk("\n->>> MEASURED CPU_FREQ %d (Hz) (%d s) <<<-\n", count, gcount/100 + 1);
		count = 0;
	}
	rt_pend_linux_irq(0);
}

int init_module(void)
{
	RESET_COUNT = SECS*100;
	rt_mount_rtai();
	rt_request_global_irq(0, fun);
	printk("\n->>> HERE WE GO (PRINTING EVERY %d SECONDS) <<<-\n\n", RESET_COUNT/100);
	return 0;
}

void cleanup_module(void)
{
	rt_free_global_irq(0);
	rt_umount_rtai();
}
