#define MODULE
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/mc146818rtc.h>

#include <asm/system.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/atomic.h>
#include <asm/desc.h>

#include <rtai.h>
#include <rtai_fifos.h>
#include <rtai_sched.h>

#define TICK_PERIOD 500000

#define STACK_SIZE 1000

#define CMDF 0

RT_TASK thread;

static int cpu_used[NR_RT_CPUS];

static void intr_handler(int t) {
	char wakeup;
	while(1) {
		cpu_used[hard_cpu_id()]++;
		rtf_put(CMDF, &wakeup, sizeof(wakeup));
		rt_task_wait_period();
	}
}


int init_module(void)
{
	RTIME tick_period;
	rtf_create_using_bh(CMDF, 4, 0);
	rt_task_init(&thread, intr_handler, 0, STACK_SIZE, 0, 0, 0);
	tick_period = start_rt_timer(nano2count(TICK_PERIOD));
	rt_task_make_periodic(&thread, rt_get_time() + 2*tick_period, tick_period);
	return 0;
}


void cleanup_module(void)
{
	int cpuid;
	stop_rt_timer();
	rt_busy_sleep(1E7);
	rt_task_delete(&thread);
	rtf_destroy(CMDF);
	printk("\n\nCPU USE SUMMARY\n");
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		printk("# %d -> %d\n", cpuid, cpu_used[cpuid]);
	}
	printk("END OF CPU USE SUMMARY\n\n");
}
