#define MODULE
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/i82489.h>
#include <asm/fixmap.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/atomic.h>
#include <asm/desc.h>

#include <rtai.h>
#include <rtai_fifos.h>
#include <rtai_sched.h>
#include <math.h>

/* the address of the parallel port -- you probably should change this */

#define ONE_SHOT

#define LPT 0x378

#define TIMER_TO_CPU 3 // < 0 || > 1 to maintain a symmetric processed timer.

#define TICK_PERIOD 20000

#define RUNNABLE_ON_CPUS 3  // 1: on cpu 0 only, 2: on cpu 1 only, 3: on any.

#define RUN_ON_CPUS (smp_num_cpus > 1 ? RUNNABLE_ON_CPUS : 1)

#define STACK_SIZE 2000

#define LOOPS 1000000000

#define PERIOD .00005

#define RTF 0

#define CMD 1

RT_TASK mytask;

static int cpu_used[NR_RT_CPUS];

void fun(int t) {
	int flag;
	double sine[2] = {1., 2.,};
	unsigned int loops = LOOPS;

	while (loops--) {
		cpu_used[hard_cpu_id()]++;
		sine[0] = (6.28/PERIOD)*((double)rt_get_time()/1193810.);
		sine[1] = sin(sine[0]);
		t = sine[1] > 0. ? 1 : 0;
		outb(t, LPT);
		rtf_put(RTF, sine, sizeof(sine));
		if (rtf_get(CMD, &flag, sizeof(flag)) == sizeof(flag)) {
			rt_task_use_fpu(rt_whoami(), flag);
		}
		rt_task_wait_period();
	}
}


int init_module(void)
{
	RTIME now, tick_period;
#ifdef ONE_SHOT
	rt_set_oneshot_mode();
#endif
	rtf_create(RTF, 4000);
	rtf_create(CMD, 100);
	rt_task_init(&mytask, fun, 0, STACK_SIZE, 0, 1, 0);
	rt_set_runnable_on_cpus(&mytask, RUN_ON_CPUS);
	tick_period = start_rt_timer((int)nano2count(TICK_PERIOD));
	rt_assign_irq_to_cpu(TIMER_8254_IRQ, TIMER_TO_CPU);
	rt_linux_use_fpu(1);
	now = rt_get_time();
	rt_task_make_periodic(&mytask, now + 2*tick_period, tick_period);
	return 0;
}


void cleanup_module(void)
{
	int cpuid;
	rt_reset_irq_to_sym_mode(TIMER_8254_IRQ);
	stop_rt_timer();
	rt_busy_sleep(1E7);
	rt_task_delete(&mytask);
	rtf_destroy(RTF);
	rtf_destroy(CMD);
	printk("\n\nCPU USE SUMMARY\n");
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		printk("# %d -> %d\n", cpuid, cpu_used[cpuid]);
	}
	printk("END OF CPU USE SUMMARY\n\n");
}
