/* MODULE CMDCLK */

#define MODULE
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <ctype.h>

#include <asm/system.h>
#include <asm/i82489.h>
#include <asm/fixmap.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/atomic.h>
#include <asm/desc.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_fifos.h>

#include "clock.h"

static enum {stopped, running} Clockstatus;
static char buffer;
static SEM notEmpty, notFull;

void CommandClock_Put(char command)
{
	rt_sem_wait(&notFull);
	if ((Clockstatus == running) == (command == 'T')) {
		buffer = command;
		rt_sem_signal(&notEmpty);
	} else {
		rt_sem_signal(&notFull);
	}
}

void CommandClock_Get(char *command)
{
	switch (Clockstatus) {
		case stopped:
			rt_sem_wait(&notEmpty);
			*command = buffer;
			if (*command == 'R') {
				Clockstatus = running;
			}
			rt_sem_signal(&notFull);
			break;
		case running:
			if (rt_sem_wait_if(&notEmpty) > 0) {
				*command = buffer;
				if (*command == 'T') {
					Clockstatus = stopped;
				}
				rt_sem_signal(&notFull);
			} else {
				*command = 'R';
			}
			break;
	}
}

int init_module(void)
{
	rt_sem_init(&notEmpty, 0);
	rt_sem_init(&notFull, 1);
	Clockstatus = stopped;
	return 0;
}

void cleanup_module(void)
{
}
