#define MODULE
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/mc146818rtc.h>

#include <asm/system.h>
#include <asm/smp.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/atomic.h>
#include <asm/desc.h>

#include <rtai.h>
#include <rtai_fifos.h>
#include <rtai_sched.h>

//#define ONESHOT

#define TIMER_TO_CPU 3 // < 0 || > 1 to maintain a symmetric processed timer.

#define DIVISOR 10

#define TICK_PERIOD ((125000 + DIVISOR/2)/DIVISOR)

#define RUNNABLE_ON_CPUS 3  // 1: on cpu 0 only, 2: on cpu 1 only, 3: on any.

#define RUN_ON_CPUS (smp_num_cpus > 1 ? RUNNABLE_ON_CPUS : 1)

#define STACK_SIZE 4000

RT_TASK thread;

static int cpu_used[NR_RT_CPUS];
static int filter(int x)
{
	static int oldx;
	int ret;

	if (x & 0x80) {
		x = 382 - x;
	}
	ret = x > oldx;
	oldx = x;
	return ret;
	
}


void intr_handler(int t) {
	char data;
	char temp;
	int divisor = DIVISOR;
	while(1) {
		if (!(--divisor)) {
			divisor = DIVISOR;
			cpu_used[hard_cpu_id()]++;
			if (rtf_get(0, &data, 1) > 0) {
				data = filter(data);
				temp = inb(0x61);            
				temp &= 0xfd;
				temp |= (data & 1) << 1;
				outb(temp,0x61);
			}
		}
		rt_task_wait_period();
	}
}


int init_module(void)
{
	RTIME now, tick_period;
	rtf_create(0, 2000);
	rt_task_init(&thread, intr_handler, 0, STACK_SIZE, 0, 0, 0);
	rt_set_runnable_on_cpus(&thread, RUN_ON_CPUS);
#ifdef ONESHOT
	rt_set_oneshot_mode();
#endif
	rt_assign_irq_to_cpu(TIMER_8254_IRQ, TIMER_TO_CPU);
	tick_period = start_rt_timer(nano2count(TICK_PERIOD));
	now = rt_get_time() + 10*tick_period;
	rt_task_make_periodic(&thread, now, tick_period);
	return 0;
}


void cleanup_module(void)
{
	int cpuid;
	rt_reset_irq_to_sym_mode(TIMER_8254_IRQ);
	stop_rt_timer();
	rt_busy_sleep(1E7);
	rt_task_delete(&thread);
	rtf_destroy(0);
	printk("\n\nCPU USE SUMMARY\n");
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		printk("# %d -> %d\n", cpuid, cpu_used[cpuid]);
	}
	printk("END OF CPU USE SUMMARY\n\n");
}
