#define MODULE
#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/io.h>

#include <rtai.h>
#include <rtai_sched.h>
#include <rtai_fifos.h>

#define DEBUG_FIFO 0

#define TIMER_TO_CPU 1 // < 0  || > 1 to maintain a symmetric processed timer.

#define RUNNABLE_ON_CPUS 2  // 1: on cpu 0 only, 2: on cpu 1 only, 3: on any;

#define RUN_ON_CPUS (smp_num_cpus > 1 ? RUNNABLE_ON_CPUS : 1)

#define SKIP 4000

#define USE_FPU 0

int period;
RTIME expected;
RT_TASK thread;
struct sample { long long min; long long max; int index; } samp;

static int cpu_used[NR_RT_CPUS];

void fun(int thread) {

	int diff;
	int skip;
	int average;
	int min_diff;
	int max_diff;

	while (1) {
		min_diff = 1000000000;
		max_diff = -1000000000;
		average = 0;

		for (skip = 0; skip < SKIP; skip++) {
			cpu_used[hard_cpu_id()]++;
			expected += period;
			rt_task_wait_period();

			diff = (int)count2nano(rt_get_time() - expected);
			if (diff < min_diff) {
				min_diff = diff;
			}
			if (diff > max_diff) {
				max_diff = diff;
			}
		average += diff;
		}
		samp.min = min_diff;
		samp.max = max_diff;
		samp.index = average/SKIP;
		rtf_put(DEBUG_FIFO, &samp, sizeof(samp));
	}
}

int init_module(void)
{
	RTIME start;
	rtf_create(DEBUG_FIFO, 16000);
	rt_linux_use_fpu(USE_FPU);
	rt_task_init(&thread, fun, 0, 3000, 0, USE_FPU, 0);
	rt_set_runnable_on_cpus(&thread, RUN_ON_CPUS);
	rt_set_oneshot_mode();
	rt_assign_irq_to_cpu(TIMER_8254_IRQ, TIMER_TO_CPU);
	period = start_rt_timer(nano2count(100000));
	start = rt_get_time() + nano2count(10000000);
	expected = start;
	rt_task_make_periodic(&thread, start, period);
	return 0;
}

void cleanup_module(void)
{
	int cpuid;
	rt_reset_irq_to_sym_mode(TIMER_8254_IRQ);
	stop_rt_timer();	
	rt_busy_sleep(1E7);
	rt_task_delete(&thread);
	rtf_destroy(DEBUG_FIFO);
	printk("\n\nCPU USE SUMMARY\n");
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		printk("# %d -> %d\n", cpuid, cpu_used[cpuid]);
	}
	printk("END OF CPU USE SUMMARY\n\n");
}
