
#define LINUX_FPU 0
#define ONE_SHOT 0
#define SCHED_IPI RTAI_4_IPI

/* This software can be used as you like but its copyright is owned by:      */
/*		Paolo Mantegazza (mantegazza@aero.polimi.it)                 */
/* It is "as it is" and I assume no responsability if anything goes wrong.   */

/* if PRIORDs below are not defined msg and sem queues are FIFOs otherwise   */
/* insertion is in priority order and receiving tasks have their priority    */
/* boosted to that of the sending task, if it is higher than their base one  */

#define NSEM_PRIORD
#define MSG_PRIORD

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/timex.h>
#include <linux/sched.h>

#include <asm/param.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/segment.h>

#include <rtai.h>
#include <rtai_sched.h>

#define RT_TASK_MAGIC 0x754d2774

#define RT_SEM_MAGIC 0xaabcdeff

#define SEM_ERR (0xFfff)

#define MSG_ERR ((RT_TASK *)0xFfff)

#define NOTHING ((void *)0)

#define SOMETHING ((void *)1)

#define __STR(x) #x
#define STR(x) __STR(x)

#define rt_exchange_tasks(oldtask, newtask) \
	__asm__ __volatile__( \
	"pushl %%eax\n\t" \
	"pushl %%ebp\n\t" \
	"pushl %%edi\n\t" \
	"pushl %%esi\n\t" \
	"pushl %%edx\n\t" \
	"pushl %%ecx\n\t" \
	"pushl %%ebx\n\t" \
	"pushl $1f\n\t" \
	"movl (%%ecx), %%ebx\n\t" \
	"movl %%esp, (%%ebx)\n\t" \
	"movl (%%edx), %%esp\n\t" \
	"movl %%edx, (%%ecx)\n\t" \
	"ret\n\t" \
"1:	popl %%ebx\n\t \
	popl %%ecx\n\t \
	popl %%edx\n\t \
	popl %%esi\n\t \
	popl %%edi\n\t \
	popl %%ebp\n\t \
	popl %%eax\n\t" \
	: /* no output */ \
	: "c" (&oldtask), "d" (newtask) \
	: "bx", "cx", "dx");

static RT_TASK *rt_smp_current[NR_RT_CPUS];

static RT_TASK *fpu_smp_task[NR_RT_CPUS];

static int linux_smp_cr0[NR_RT_CPUS];

static RT_TASK rt_smp_linux_task[NR_RT_CPUS];

#define DECLARE_RT_CURRENT RT_TASK *rt_current

#define RT_CURRENT rt_smp_current[hard_cpu_id()]

#define ASSIGN_RT_CURRENT rt_current = rt_smp_current[hard_cpu_id()]

#define fpu_task (fpu_smp_task[cpuid])

#define linux_cr0 (linux_smp_cr0[cpuid])

#define rt_linux_task (rt_smp_linux_task[cpuid])

#define rt_base_linux_task (rt_smp_linux_task[NR_RT_CPUS - 1])

static RTIME rt_time_h;

static int rt_half_tick;

static int oneshot_timer;

static int not_fired;

static int preempt_always;

static spinlock_t oneshot_lock;

extern unsigned long cpu_present_map;

/* ++++++++++++++++++++++++++++++++ TASKS ++++++++++++++++++++++++++++++++++ */

extern unsigned int *locked_cpus;

static inline void rt_get_global_lock(int cpuid)
{
	if (!test_and_set_bit(cpuid, locked_cpus)) {
		while (test_and_set_bit(31, locked_cpus));
	}
}


static inline void rt_release_global_lock(int cpuid)
{
	if (test_and_clear_bit(cpuid, locked_cpus)) {
		clear_bit(31, locked_cpus);
	}
}


static void rt_startup(void (*rt_thread)(int), int data)
{
	DECLARE_RT_CURRENT;
	ASSIGN_RT_CURRENT;
	rt_global_sti();
	(*rt_thread)(data);
	rt_task_delete(rt_current);
}


int rt_task_init(RT_TASK *task, void (*rt_thread)(int), int data,
			int stack_size, int priority, int uses_fpu,
			void(*signal)(void))
{
	int *st;
	unsigned long flags;
	int cpuid;

	if (task->magic == RT_TASK_MAGIC) {
		return -EINVAL;
	}

	st = (int *) kmalloc(stack_size, GFP_KERNEL);   
	if (!st) {
		return -ENOMEM;
	}

	task->stack = st + stack_size / sizeof(int);    
	task->uses_fpu = uses_fpu;
	task->runnable_on_cpus = cpu_present_map;
	task->stack_bottom = st;
	task->magic = RT_TASK_MAGIC; 
	task->state = (SUSPENDED | READY);
	task->priority = task->base_priority = priority;
	task->period = 0;
	task->resume_time = RT_TIME_END;
	task->queue.prev = &(task->queue);      
	task->queue.next = &(task->queue);      
	task->queue.task = task;
	task->msg_queue.prev = &(task->msg_queue);      
	task->msg_queue.next = &(task->msg_queue);      
	task->msg_queue.task = task;    
	task->msg = 0;  
	task->blocked_on.task = NOTHING;        
	task->signal = signal;

	*--(task->stack) = data;
	*--(task->stack) = (int) rt_thread;
	*--(task->stack) = 0;
	*--(task->stack) = (int) rt_startup;

	flags = rt_global_save_flags_and_cli();
	cpuid = hard_cpu_id();
	save_cr0_and_clts(linux_cr0);
	save_fpenv(fpu_task->fpu_reg);
	save_fpenv(task->fpu_reg);
	restore_fpenv(fpu_task->fpu_reg);
	restore_cr0(linux_cr0);
	rt_base_linux_task.prev->next = task;
	task->prev = rt_base_linux_task.prev;
	task->next = 0;
	rt_base_linux_task.prev = task;
	rt_global_restore_flags(flags);

	return 0;
}


void rt_set_runnable_on_cpus(RT_TASK *task, unsigned int runnable_on_cpus)
{
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	if (!(task->runnable_on_cpus = (runnable_on_cpus & cpu_present_map))) {
		task->runnable_on_cpus = 1;
	}
	rt_global_restore_flags(flags);
}


static void rt_schedule_on_schedule_ipi(void)
{
	DECLARE_RT_CURRENT;
	RT_TASK *task, *new_task;
	int prio, cpuid;

	prio = RT_LOWEST_PRIORITY;
	task = &rt_base_linux_task;
	rt_current = rt_smp_current[cpuid = hard_cpu_id()];
	new_task = &rt_linux_task;

	rt_get_global_lock(cpuid);
	rt_current->state &= ~RUNNING;
	while ((task = task->next)) {
		if (task->state == READY &&
		    test_bit(cpuid, &(task->runnable_on_cpus)) &&
		    task->priority < prio) { 
			new_task = task;
			prio = task->priority;
		}
	}
	new_task->state = (READY | RUNNING);
	rt_release_global_lock(cpuid);

	if (new_task != rt_current) {
		if (rt_current == &rt_linux_task) {
			rt_switch_to_real_time(cpuid);
			save_cr0_and_clts(linux_cr0);
		}
		if (new_task->uses_fpu) {
			if (new_task != fpu_task) {
				save_fpenv(fpu_task->fpu_reg);
				fpu_task = new_task;
				restore_fpenv(fpu_task->fpu_reg);
			}
		}
		rt_exchange_tasks(rt_smp_current[cpuid], new_task);
		if (rt_current->signal) {
			(*rt_current->signal)();
		}
	}
}


static void rt_schedule(void)
{
	DECLARE_RT_CURRENT;
	RT_TASK *task, *new_task;
	unsigned int cpus_with_ready_tasks;
	int prio, delay, preemptor, cpuid;

	task = &rt_base_linux_task;
	cpus_with_ready_tasks = 0;
	prio = RT_LOWEST_PRIORITY;
	rt_current = rt_smp_current[cpuid = hard_cpu_id()];
	new_task = &rt_linux_task;
	rt_current->state &= ~RUNNING;
	if (oneshot_timer) {
		rt_time_h = rdtsc() + rt_half_tick;
		while ((task = task->next)) {
			if ((task->state & DELAYED) && task->resume_time <= rt_time_h) {
				if (task->state & (SEMAPHORE | SEND | RPC)) {
					(task->queue.prev)->next = task->queue.next;
					(task->queue.next)->prev = task->queue.prev;
				}
				task->state &= ~(DELAYED | SEMAPHORE | RECEIVE |
						 SEND    | RPC       | RETURN);
			}
			if (task->state == READY) {
				if (test_bit(cpuid, &(task->runnable_on_cpus)) &&
			    	    task->priority < prio) { 
					cpus_with_ready_tasks |= new_task->runnable_on_cpus;
					new_task = task;
					prio = task->priority;
				} else {
					cpus_with_ready_tasks |= task->runnable_on_cpus;
				}
			}
		}
		new_task->state = (READY | RUNNING);
		send_ipi_logical(cpus_with_ready_tasks & ~(1 << cpuid), SCHED_IPI);
		rt_release_global_lock(cpuid);
		rt_spin_lock(&oneshot_lock);
		preemptor = 0;
		task = &rt_base_linux_task;
		while ((task = task->next)) {
			if ((task->state & DELAYED) && task->priority <= prio && task->resume_time < rt_times.intr_time) {
				rt_times.intr_time = task->resume_time;
				preemptor = 1;
			}
		}
		if (preemptor || (not_fired && prio == RT_LOWEST_PRIORITY)) {
			not_fired = 0;
			delay = (int)(rt_times.intr_time - rdtsc()) - tuned.latency;
			if (delay >= tuned.setup_time_8254CPU) {
				delay = imuldiv(delay, FREQ_8254, tuned.cpu_freq);
			} else {
				delay = tuned.setup_time_8254;
				rt_times.intr_time = rdtsc() + (tuned.setup_time_8254CPU);
			}
			outb(delay & 0xFF, 0x40);
			outb(delay >> 8, 0x40);
		}
		rt_spin_unlock(&oneshot_lock);
	} else {
		while ((task = task->next)) {
			if (task->state == READY) {
				if (test_bit(cpuid, &(task->runnable_on_cpus)) &&
			    	    task->priority < prio) { 
					cpus_with_ready_tasks |= new_task->runnable_on_cpus;
					new_task = task;
					prio = task->priority;
				} else {
					cpus_with_ready_tasks |= task->runnable_on_cpus;
				}
			}
		}
		new_task->state = (READY | RUNNING);
		send_ipi_logical(cpus_with_ready_tasks & ~(1 << cpuid), SCHED_IPI);
		rt_release_global_lock(cpuid);
	}

	if (new_task != rt_current) {
		if (rt_current == &rt_linux_task) {
			rt_switch_to_real_time(cpuid);
			save_cr0_and_clts(linux_cr0);
		}
		if (new_task->uses_fpu) {
			if (new_task != fpu_task) {
				save_fpenv(fpu_task->fpu_reg);
				fpu_task = new_task;
				restore_fpenv(fpu_task->fpu_reg);
			}
		}
		if (new_task == &rt_linux_task) {
			rt_switch_to_linux(cpuid);
			restore_cr0(linux_cr0);
		}
		rt_exchange_tasks(rt_smp_current[cpuid], new_task);
		if (rt_current->signal) {
			(*rt_current->signal)();
		}
	}
	rt_get_global_lock(hard_cpu_id());
}


int rt_task_delete(RT_TASK *task)
{
	int cpuid;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	flags = rt_global_save_flags_and_cli();
	if (task->state & SEMAPHORE) {
		(task->blocked_on.sem)->count++;
		(task->queue.prev)->next = task->queue.next;
		(task->queue.next)->prev = task->queue.prev;
	} else if (task->state & (SEND | RPC)) {
		(task->queue.prev)->next = task->queue.next;
		(task->queue.next)->prev = task->queue.prev;
	} else {
		q = &(task->msg_queue);
		while ((q = q->next) != &(task->msg_queue)) {
			(q->task)->state &= ~(SEND | RPC | DELAYED);
		}       
	}
	task->magic = 0;
	if (!((task->prev)->next = task->next)) {
		rt_base_linux_task.prev = task->prev;
	} else {
		(task->next)->prev = task->prev;
	}
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		if (fpu_task == task) {
			fpu_task = &rt_linux_task;
			break;
		}
	}
	kfree(task->stack_bottom);
	if (task == RT_CURRENT) {
		rt_schedule();
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_get_task_state(RT_TASK *task)
{
	return task->state;
}


static void rt_timer_handler(void)
{
	DECLARE_RT_CURRENT;
	RT_TASK *task, *new_task;
	unsigned int cpus_with_ready_tasks;
	int prio, delay, preemptor, cpuid; 

	task = &rt_base_linux_task;
	prio = RT_LOWEST_PRIORITY;
	cpus_with_ready_tasks = 0;
	rt_current = rt_smp_current[cpuid = hard_cpu_id()];
	new_task = &rt_linux_task;

	rt_get_global_lock(cpuid);
	rt_times.tick_time = rt_times.intr_time;
	rt_time_h = rt_times.tick_time + rt_half_tick;
	if (rt_times.tick_time >= rt_times.linux_time) {
		rt_times.linux_time += rt_times.linux_tick;
		rt_pend_linux_irq(0);
	}
	rt_current->state &= ~RUNNING;
	while ((task = task->next)) {
		if ((task->state & DELAYED) && task->resume_time <= rt_time_h) {
			if (task->state & (SEMAPHORE | SEND | RPC)) {
				(task->queue.prev)->next = task->queue.next;
				(task->queue.next)->prev = task->queue.prev;
			}
			task->state &= ~(DELAYED | SEMAPHORE | RECEIVE |
					 SEND    | RPC       | RETURN);
		}
		if (task->state == READY) {
			if (test_bit(cpuid, &(task->runnable_on_cpus)) &&
		    	    task->priority < prio) { 
				cpus_with_ready_tasks |= new_task->runnable_on_cpus;
				new_task = task;
				prio = task->priority;
			} else {
				cpus_with_ready_tasks |= task->runnable_on_cpus;
			}
		}
	}
	new_task->state = (READY | RUNNING);
	send_ipi_logical(cpus_with_ready_tasks & ~(1 << cpuid), SCHED_IPI);

	if (oneshot_timer) {
		rt_times.intr_time = rt_times.linux_time > rt_times.tick_time ?
		rt_times.linux_time : rt_times.tick_time + (rt_times.linux_tick>>1);
		rt_release_global_lock(cpuid);
		rt_spin_lock(&oneshot_lock);
		preemptor = preempt_always;
		task = &rt_base_linux_task;
		while ((task = task->next)) {
			if ((task->state & DELAYED) && task->priority <= prio &&
				      task->resume_time < rt_times.intr_time) {
				rt_times.intr_time = task->resume_time;
				preemptor = 1;
			}
		}
		if (preemptor || prio == RT_LOWEST_PRIORITY) {
			not_fired = 0;
			delay = (int)(rt_times.intr_time - rdtsc()) - tuned.latency;
			if (delay >= tuned.setup_time_8254CPU) {
				delay = imuldiv(delay, FREQ_8254, tuned.cpu_freq);
			} else {
				delay = tuned.setup_time_8254;
				rt_times.intr_time = rdtsc() + (tuned.setup_time_8254CPU);
			}
			outb(delay & 0xFF, 0x40);
			outb(delay >> 8, 0x40);
		} else {
			not_fired = 1;
		}
		rt_spin_unlock(&oneshot_lock);
	} else {
		rt_times.intr_time += rt_times.periodic_tick;
		rt_release_global_lock(cpuid);
	}

	if (new_task != rt_current) {
		if (rt_current == &rt_linux_task) {
			rt_switch_to_real_time(cpuid);
			save_cr0_and_clts(linux_cr0);
		}
		if (new_task->uses_fpu) {
			if (new_task != fpu_task) {
				save_fpenv(fpu_task->fpu_reg);
				fpu_task = new_task;
				restore_fpenv(fpu_task->fpu_reg);
			}
		}
		rt_exchange_tasks(rt_smp_current[cpuid], new_task);
		if (rt_current->signal) {
			(*rt_current->signal)();
		}
	}
}


static void recover_jiffies(int irq, void *dev_id, struct pt_regs *regs)
{
	rt_global_cli();
	if (rt_times.tick_time >= rt_times.linux_time) {
		rt_times.linux_time += rt_times.linux_tick;
		rt_pend_linux_irq(0);
	}
	rt_global_sti();
} 


void rt_set_periodic_mode(void)
{
	stop_rt_timer();
	oneshot_timer = 0;
}


void rt_set_oneshot_mode(void)
{
	stop_rt_timer();
	oneshot_timer = 1;
}


RTIME start_rt_timer(int period)
{
	unsigned long flags;

	flags = rt_global_save_flags_and_cli();
	if (oneshot_timer) {
#if CPU_FREQ == FREQ_8254
		rt_setup_8254_tsc();
#endif
		rt_request_timer(rt_timer_handler, 0);
		rt_half_tick = tuned.latency;
		not_fired = 0;
	} else {
		rt_request_timer(rt_timer_handler, period > LATCH? LATCH: period);
		rt_half_tick = (rt_times.periodic_tick + 1)>>1;
	}
	rt_time_h = rt_times.tick_time + rt_half_tick;
	rt_request_linux_irq(TIMER_8254_IRQ, recover_jiffies, "rtai_sched", recover_jiffies);
	rt_request_cpu_own_irq(SCHED_IPI, rt_schedule_on_schedule_ipi);
	rt_global_restore_flags(flags);
	return period;
}


void stop_rt_timer(void)
{
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	rt_free_timer();
	rt_free_linux_irq(TIMER_8254_IRQ, recover_jiffies);
	oneshot_timer = 0;
	rt_global_restore_flags(flags);
	rt_busy_sleep(10000);
	rt_free_cpu_own_irq(SCHED_IPI);
}


RT_TASK *rt_whoami(void)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	rt_global_restore_flags(flags);
	return rt_current;
}


int rt_task_signal_handler(RT_TASK *task, void (*handler)(void))
{
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	task->signal = handler;
	return 0;
}


int rt_task_use_fpu(RT_TASK *task, int use_fpu_flag)
{
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	task->uses_fpu = use_fpu_flag != 0 ? 1 : 0;
	return 0;
}


void rt_linux_use_fpu(int use_fpu_flag)
{
	int cpuid;
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		rt_linux_task.uses_fpu = use_fpu_flag != 0 ? 1 : 0;
	}
}


void rt_preempt_always(int yes_no)
{
	if (oneshot_timer) {
		preempt_always = yes_no != 0 ? 1 : 0;
	}
}


const int OneShot = ONE_SHOT;
MODULE_PARM(OneShot, "i");

const int LinuxFpu = LINUX_FPU;
MODULE_PARM(LinuxFpu, "i");

const int CpuFreq = CPU_FREQ;
MODULE_PARM(CpuFreq, "i");

const int Latency = LATENCY;
MODULE_PARM(Latency, "i");

const int SetupTime8254 = SETUP_TIME_8254;
MODULE_PARM(SetupTime8254, "i");

int init_module(void)
{
	int cpuid;
	for (cpuid = 0; cpuid < NR_RT_CPUS; cpuid++) {
		rt_linux_task.uses_fpu = LinuxFpu != 0 ? 1 : 0;
		rt_linux_task.magic = 0;
		rt_linux_task.state = (READY | RUNNING);
		rt_linux_task.priority = RT_LOWEST_PRIORITY;
		rt_linux_task.base_priority = RT_LOWEST_PRIORITY;
		rt_linux_task.signal = 0;
		rt_linux_task.prev = &rt_base_linux_task;
		rt_linux_task.next = 0;
		rt_smp_current[cpuid] = &rt_linux_task;
		fpu_smp_task[cpuid] = &rt_linux_task;
	}
	tuned.cpu_freq = CpuFreq;
	tuned.latency = imuldiv(Latency, CpuFreq, 1E9);
	tuned.setup_time_8254CPU = imuldiv(SetupTime8254, CpuFreq, 1E9);
	tuned.setup_time_8254 = imuldiv(SetupTime8254, FREQ_8254, 1E9);
	oneshot_timer = OneShot != 0 ? 1 : 0;
	preempt_always = 0;
	printk("\n***** STARTING THE REAL TIME SCHEDULER WITH %sLINUX *****", rt_base_linux_task.uses_fpu ? "": "NO ");
	printk("\n***** FP SUPPORT AND READY FOR A %s TIMER *****", oneshot_timer ? "ONESHOT": "PERIODIC");
	printk("\n***<> LINUX TICK AT %d (HZ) <>***", HZ);
	printk("\n***<> CALIBRATED CPU FREQUENCY %d (HZ) <>***", tuned.cpu_freq);
	printk("\n***<> CALIBRATED TIMER-INTERRUPT-TO-SCHEDULER LATENCY %d (ns) <>***", imuldiv(tuned.latency - tuned.setup_time_8254CPU, 1E9, tuned.cpu_freq));
	printk("\n***<> CALIBRATED ONE SHOT SETUP TIME %d (ns) <>***\n\n", imuldiv(tuned.setup_time_8254CPU, 1E9, tuned.cpu_freq));
	rt_mount_rtai();
	return 0;
}


void cleanup_module(void)
{
	stop_rt_timer();
	while (rt_base_linux_task.next) {
		rt_task_delete(rt_base_linux_task.next);
	}
	printk("\n***** THE REAL TIME SCHEDULER HAS BEEN REMOVED *****\n\n");
	rt_umount_rtai();
}

/* ++++++++++++++++++++++++++ TIME CONVERSIONS +++++++++++++++++++++++++++++ */

RTIME count2nano(RTIME counts)
{
	int sign;

	if (counts > 0) {
		sign = 1;
	} else {
		sign = 0;
		counts = - counts;
	}
	counts = oneshot_timer ?
		 llimd(counts, 1E9, tuned.cpu_freq):
		 llimd(counts, 1E9, FREQ_8254);
	return sign ? counts : - counts;
}


RTIME nano2count(RTIME ns)
{
	int sign;

	if (ns > 0) {
		sign = 1;
	} else {
		sign = 0;
		ns = - ns;
	}
	ns =  oneshot_timer ?
	      llimd(ns, tuned.cpu_freq, 1E9) :
	      llimd(ns, FREQ_8254, 1E9);
	return sign ? ns : - ns;
}

/* +++++++++++++++++++++++++++++++ TIMINGS ++++++++++++++++++++++++++++++++++ */

RTIME rt_get_time(void)
{
	return oneshot_timer ? rdtsc(): rt_times.tick_time;
}


RTIME rt_get_time_ns(void)
{
	return oneshot_timer ?
	       llimd(rdtsc(), 1E9, tuned.cpu_freq) :
	       llimd(rt_times.tick_time, 1E9, FREQ_8254);
}


RTIME rt_get_cpu_time_ns(void)
{
	return llimd(rdtsc(), 1E9, tuned.cpu_freq);
}


void rt_task_yield(void)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (rt_current->next) {
		(rt_current->prev)->next = rt_current->next;
		(rt_current->next)->prev = rt_current->prev;
		rt_base_linux_task.prev->next = rt_current;
		rt_current->prev = rt_base_linux_task.prev;
		rt_current->next = 0;
		rt_base_linux_task.prev = rt_current;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}


int rt_task_suspend(RT_TASK *task)
{
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	flags = rt_global_save_flags_and_cli();
	task->state |= SUSPENDED;
	if (task == RT_CURRENT) {
		rt_schedule();
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_task_resume(RT_TASK *task)
{
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}

	flags = rt_global_save_flags_and_cli();
	if (((task->state &= ~SUSPENDED) & ~DELAYED) == READY) {
		rt_schedule();
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_task_make_periodic(RT_TASK *task, RTIME start_time, RTIME period)
{
	long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	flags = rt_global_save_flags_and_cli();
	task->resume_time = start_time;
	task->period = period;
	task->state = (task->state &~SUSPENDED) | DELAYED;
	rt_schedule();
	rt_global_restore_flags(flags);
	return 0;
}


void rt_task_wait_period(void)
{
	DECLARE_RT_CURRENT;
	long flags;

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((rt_current->resume_time += rt_current->period) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}


RTIME next_period(void)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	rt_global_restore_flags(flags);
	return rt_current->resume_time + rt_current->period;
}


void rt_busy_sleep(int ns)
{
	RTIME end_time;
	end_time = rdtsc() + llimd(ns, tuned.cpu_freq, 1E9);
	while (rdtsc() < end_time);
}


void rt_sleep(RTIME delay)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((rt_current->resume_time = (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}


void rt_sleep_until(RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((rt_current->resume_time = time) > rt_time_h) {
		rt_current->state |= DELAYED;
		rt_schedule();
	}
	rt_global_restore_flags(flags);
}

/* +++++++++++++++++++++++++++++ SEMAPHORES ++++++++++++++++++++++++++++++++ */

void rt_sem_init(SEM *sem, int value)
{
	sem->count = value;
	sem->magic = RT_SEM_MAGIC;
	sem->queue.prev = &(sem->queue);
	sem->queue.next = &(sem->queue);
	sem->queue.task = 0;
}


int rt_sem_delete(SEM *sem)
{
	unsigned long flags;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	sem->magic = 0;
	q = &(sem->queue);
	while ((q = q->next) != &(sem->queue)) {
		(q->task)->state &= ~(SEMAPHORE | DELAYED);
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_sem_signal(SEM *sem)
{
	unsigned long flags;
	RT_TASK *task;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	if ((sem->count)++ < 0) {
		if ((task = (sem->queue.next)->task)) {
			sem->queue.next = task->queue.next;
			(task->queue.next)->prev = &(sem->queue);
			task->blocked_on.sem = NOTHING;
			if ((task->state &= ~(SEMAPHORE | DELAYED)) == READY) {
				rt_schedule();
			}
		}
	}
	rt_global_restore_flags(flags);
	return 0;
}


int rt_sem_wait(SEM *sem)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	int count;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	count = sem->count;
	if (--(sem->count) < 0) {
		rt_current->state |= SEMAPHORE;
		rt_current->blocked_on.sem = sem;
		q = &(sem->queue);
#ifdef SEM_PRIORD
		while ((q = q->next) != &(sem->queue) &&
			 (q->task)->priority <= rt_current->priority);
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_schedule();
		if (rt_current->blocked_on.sem || sem->magic != RT_SEM_MAGIC) {
			count = SEM_ERR;
		} else { 
			count = sem->count;
		}
	}
	rt_global_restore_flags(flags);
	return count;
}


int rt_sem_wait_if(SEM *sem)
{
	int count;
	unsigned long flags;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	if ((count = sem->count) > 0) {
		sem->count--;
	}
	rt_global_restore_flags(flags);
	return count;
}


int rt_sem_wait_until(SEM *sem, RTIME time)
{
	DECLARE_RT_CURRENT;
	int count;
	unsigned long flags;
	QUEUE *q;

	if (sem->magic != RT_SEM_MAGIC) {
		return SEM_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	count = sem->count;
	if (--(sem->count) < 0) {
		rt_current->blocked_on.sem = sem;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= (SEMAPHORE | DELAYED);
			q = &(sem->queue);
#ifdef SEM_PRIORD
			while ((q = q->next) != &(sem->queue) &&
				 (q->task)->priority <= rt_current->priority);
#endif
			rt_current->queue.prev = q->prev;
			rt_current->queue.next = q;
			(q->prev)->next = &(rt_current->queue);
			q->prev = &(rt_current->queue);
			rt_schedule();
		}
		if (sem->magic != RT_SEM_MAGIC) {
			count = SEM_ERR;
		} else {
			if (rt_current->blocked_on.sem) {
				rt_current->blocked_on.sem = NOTHING;
				count = sem->count;
				sem->count++;
			} else {
				count = sem->count;
			}
		}
	}
	rt_global_restore_flags(flags);
	return count;
}


int rt_sem_wait_timed(SEM *sem, RTIME delay)
{
	return rt_sem_wait_until(sem, (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay);
}

/* ++++++++++++++++++++++++++++++ MESSAGES +++++++++++++++++++++++++++++++++ */

/* ++++++++++++++++++++++++++++++++ SEND +++++++++++++++++++++++++++++++++++ */

RT_TASK *rt_send(RT_TASK *task, unsigned int msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		rt_current->msg = msg;
		rt_current->msg_queue.task = task;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= SEND;
		rt_schedule();
	}
	if (rt_current->msg_queue.task != rt_current) {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_if(RT_TASK *task, unsigned int msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			rt_schedule();
		}
		if (rt_current->msg_queue.task != rt_current) {
			rt_current->msg_queue.task = rt_current;
			task = (RT_TASK *)0;
		}
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_until(RT_TASK *task, unsigned int msg, RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = msg;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		if ((task->state &= ~(RECEIVE | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		rt_current->msg_queue.task = task;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->msg = msg;
			q = &(task->msg_queue);
#ifdef MSG_PRIORD
			while ((q = q->next) != &(task->msg_queue) &&
				 (q->task)->priority <= rt_current->priority);
			if (task->priority > rt_current->priority) {
				task->priority = rt_current->priority;
			}
#endif
			rt_current->queue.prev = q->prev;
			rt_current->queue.next = q;
			(q->prev)->next = &(rt_current->queue);
			q->prev = &(rt_current->queue);
			rt_current->state |= (SEND | DELAYED);
			rt_schedule();
		}
	}
	if (rt_current->msg_queue.task != rt_current) {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_send_timed(RT_TASK *task, unsigned int msg, RTIME delay)
{
	return rt_send_until(task, msg, (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay);
}

/* +++++++++++++++++++++++++++++++++ RPC +++++++++++++++++++++++++++++++++++ */

RT_TASK *rt_rpc(RT_TASK *task, unsigned int to_do, unsigned int *result)
{

	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
		(!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= RETURN;
	} else {
		rt_current->msg = to_do;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= RPC;
	}
	rt_current->msg_queue.task = task;
	rt_schedule();
	if (rt_current->msg_queue.task == rt_current) {
		*result = rt_current->msg;
	} else {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_if(RT_TASK *task, unsigned int to_do, unsigned int *result)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}
	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= RETURN;
		rt_current->msg_queue.task = task;
		rt_schedule();
		if (rt_current->msg_queue.task == rt_current) {
			*result = rt_current->msg;
		} else {
			rt_current->msg_queue.task = rt_current;
			task = (RT_TASK *)0;
		}
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_until(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;
	QUEUE *q;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RECEIVE) &&
	    (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
		task->msg = to_do;
		task->msg_queue.task = rt_current;
		task->blocked_on.task = NOTHING;
		task->state &= ~(RECEIVE | DELAYED);
		rt_current->state |= (RETURN | DELAYED);
	} else {
		if ((rt_current->resume_time = time) <= rt_time_h) {
			rt_global_restore_flags(flags);
			return (RT_TASK *)0;
		}
		rt_current->msg = to_do;
		q = &(task->msg_queue);
#ifdef MSG_PRIORD
		while ((q = q->next) != &(task->msg_queue) &&
			 (q->task)->priority <= rt_current->priority);
		if (task->priority > rt_current->priority) {
			task->priority = rt_current->priority;
		}
#endif
		rt_current->queue.prev = q->prev;
		rt_current->queue.next = q;
		(q->prev)->next = &(rt_current->queue);
		q->prev = &(rt_current->queue);
		rt_current->state |= (RPC | DELAYED);
	}
	rt_current->msg_queue.task = task;
	rt_schedule();
	if (rt_current->msg_queue.task == rt_current) {
		*result = rt_current->msg;
	} else {
		rt_current->msg_queue.task = rt_current;
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_rpc_timed(RT_TASK *task, unsigned int to_do, unsigned int *result, RTIME delay)
{
	return rt_rpc_until(task, to_do, result, (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay);
}

/* ++++++++++++++++++++++++++++++ RPC_RETURN +++++++++++++++++++++++++++++++ */

int rt_isrpc(RT_TASK *task)
{
	return task->state & RETURN;
}


RT_TASK *rt_return(RT_TASK *task, unsigned int result)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if ((task->state & RETURN) && task->msg_queue.task == rt_current) {
		task->msg = result;
		task->msg_queue.task = task;
		if ((task->state &= ~(RETURN | DELAYED)) == READY) {
			rt_schedule();
		}
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}

/* +++++++++++++++++++++++++++++++ RECEIVE +++++++++++++++++++++++++++++++++ */

RT_TASK *rt_receive(RT_TASK *task, unsigned int *msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		rt_current->msg_queue.task = task;
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				rt_schedule();
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
	} else {
#ifdef MSG_PRIORD
		task->priority = rt_current->base_priority;
#endif
		rt_current->blocked_on.task = SOMETHING;
		rt_current->state |= RECEIVE;
		rt_current->msg_queue.task = (task != rt_current) ? task : (RT_TASK *)0;
		rt_schedule();
		*msg = rt_current->msg;
	}
	if (rt_current->blocked_on.task) {
		rt_current->blocked_on.task = NOTHING;
		task = (RT_TASK *)0;
	} else {
		task = rt_current->msg_queue.task;
	}
	rt_current->msg_queue.task = rt_current;
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_if(RT_TASK *task, unsigned int *msg)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		rt_current->msg_queue.task = task;
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				rt_schedule();
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
		if (rt_current->blocked_on.task) {
			rt_current->blocked_on.task = NOTHING;
			task = (RT_TASK *)0;
		} else {
			task = rt_current->msg_queue.task;
		}
		rt_current->msg_queue.task = rt_current;
	} else {
		task = (RT_TASK *)0;
	}
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_until(RT_TASK *task, unsigned int *msg, RTIME time)
{
	DECLARE_RT_CURRENT;
	unsigned long flags;

	if (task && task->magic != RT_TASK_MAGIC) {
		return MSG_ERR;
	}

	flags = rt_global_save_flags_and_cli();
	ASSIGN_RT_CURRENT;
	if (!task) task = (rt_current->msg_queue.next)->task;
	if ((task->state & (SEND | RPC)) && task->msg_queue.task == rt_current) {
		(task->queue.next)->prev = task->queue.prev;
		(task->queue.prev)->next = task->queue.next;
		*msg = task->msg;
#ifdef MSG_PRIORD
		rt_current->priority = rt_current->base_priority > task->priority ?
			task->priority : rt_current->base_priority;
#endif
		rt_current->msg_queue.task = task;
		if (task->state & SEND) {
			task->msg_queue.task = task;
			if ((task->state  &=  ~(SEND | DELAYED)) == READY) {
				rt_schedule();
			}
		} else if (task->state & RPC) {
			task->state = (task->state & ~(RPC | DELAYED)) | RETURN;
		}
	} else {
#ifdef MSG_PRIORD
		task->priority = rt_current->base_priority;
#endif
		rt_current->blocked_on.task = SOMETHING;
		if ((rt_current->resume_time = time) > rt_time_h) {
			rt_current->state |= (RECEIVE | DELAYED);
			rt_current->msg_queue.task = (task != rt_current) ? task : (RT_TASK *)0;
			rt_schedule();
			*msg = rt_current->msg;
		}
	}
	if (rt_current->blocked_on.task) {
		rt_current->blocked_on.task = NOTHING;
		task = (RT_TASK *)0;
	} else {
		task = rt_current->msg_queue.task;
	}
	rt_current->msg_queue.task = rt_current;
	rt_global_restore_flags(flags);
	return task;
}


RT_TASK *rt_receive_timed(RT_TASK *task, unsigned int *msg, RTIME delay)
{
	return rt_receive_until(task, msg, (oneshot_timer ? rdtsc(): rt_times.tick_time) + delay);
}
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
