
#define FREQ_8254 1193180
//#define CPU_FREQ 1193180
//#define CPU_FREQ 233863280
#define CPU_FREQ 350794919
#define LATENCY 9500
#define SETUP_TIME_8254 2500

#define TIMER_8254_IRQ       0x00

#define RTAI_1_IPI      6
#define RTAI_2_IPI      7
#define RTAI_3_IPI      8
#define RTAI_4_IPI      9

#define RTAI_1_VECTOR         0xD9
#define RTAI_2_VECTOR         0xE1
#define RTAI_3_VECTOR         0xE9
#define RTAI_4_VECTOR         0xF1

extern void rt_global_cli(void);
extern void rt_global_sti(void);
extern void rt_global_save_flags(unsigned long *flags);
extern void rt_global_restore_flags(unsigned long flags);
extern int rt_global_save_flags_and_cli(void);
extern int hard_lock_all(void);
extern void hard_unlock_all(unsigned long flags);
extern void send_ipi_shorthand(unsigned int shorthand, int irq);
extern void send_ipi_logical(unsigned long dest, int irq);
extern int rt_assign_irq_to_cpu(int irq, int cpu);
extern int rt_reset_irq_to_sym_mode(int irq);
extern int  rt_request_global_irq(unsigned int irq, void (*handler)(void));
extern int  rt_free_global_irq(unsigned int irq);
extern void rt_enable_irq(unsigned int irq);
extern void rt_disable_irq(unsigned int irq);
extern int rt_request_linux_irq(unsigned int irq,
	void (*linux_handler)(int irq, void *dev_id, struct pt_regs *regs), 
	char *linux_handler_id, void *dev_id);
extern int rt_free_linux_irq(unsigned int irq, void *dev_id);
extern void rt_pend_linux_irq(unsigned int irq);
extern int rt_request_srq(void (*handler)(void));
extern int rt_free_srq(unsigned int srq);
extern void rt_pend_linux_srq(unsigned int srq);
extern int rt_request_cpu_own_irq(unsigned int irq, void (*handler)(void));
extern int rt_free_cpu_own_irq(unsigned int irq);
extern void rt_switch_to_linux(int cpuid);
extern void rt_switch_to_real_time(int cpuid);
extern void rt_request_timer(void (*handler)(void), int tick);
extern void rt_free_timer(void);
extern void rt_mount_rtai(void);
extern void rt_umount_rtai(void);

// defines for compatibility with 2.0.xx RTLinux
#define request_RTirq(irq, handler)  rt_request_global_irq(irq, handler)
#define free_RTirq(irq)              rt_free_global_irq(irq)
#define disable_RTirq(irq)           rt_disable_irq(irq)
#define enable_RTirq(irq)            rt_enable_irq(irq)

#ifdef __SMP__

// defines for compatibility with 2.0.xx RTLinux
#define r_cli()                      rt_global_cli()
#define r_sti()                      rt_global_sti()
#define r_save_flags(flags)          rt_global_save_flags(&flags)
#define r_restore_flags(flags)       rt_global_restore_flags(flags)

static volatile inline void rt_spin_lock(spinlock_t *lock)
{
	while(test_and_set_bit(0, &(lock->lock)));
}

static volatile inline void rt_spin_unlock(spinlock_t *lock)
{
 	clear_bit(0, &(lock->lock));
}

#define hard_cpu_id()		     hard_smp_processor_id() 

//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// If NR_RT_CPUS > 8 RTAI must be changed as it cannot use APIC flat delivery
// and the way processor[?].intr_flag is used now must be changed (right now
// it exploits the fact that the IF flags is at bit 9 so that bits 0-7 are used
// to mark a cpu as Linux soft irq enabled/disabled. Bad but comfortable, it 
// will take a very very long time before I'll have available SMP with more
// than 8 cpus. Right now they are only:
#define NR_RT_CPUS   2

#else

// defines for compatibility with 2.0.xx RTLinux
#define r_cli()                      hard_cli()
#define r_sti()                      hard_sti()
#define r_save_flags(flags)          hard_save_flags(flags)
#define r_restore_flags(flags)       hard_restore_flags(flags)

#define rt_spin_lock(whatever)  
#define rt_spin_unlock(whatever)

#define hard_cpu_id()                0

#define NR_RT_CPUS   1

#endif

// macros to support hard floating point in interrupt handlers:
// typical sequence of use:
// save_cr0_and_clts; 	#to save the Linux cr0 statei, always to be done.
// save_fpenv; 		# to save Linux fpu environment, needed only any Linux
//			  process uses the fpu.
// restore_fpenv;	# to restore your fou environment, needed only if it
//			# can be interrupted or if you left some intermediate
//			# results.
// save_fpenv;		# to save your fpu environment, only if you leave with
//			# intermediate results to be used afterward.
// restore_fpenv;	# to restore Linux fpu environment.
// restore_cr0;		# to restore Linux cr0, always to be done.
		 
#define save_cr0_and_clts(x) __asm__ __volatile__ ("movl %%cr0,%0; clts" :"=r" (x))

#define restore_cr0(x)   __asm__ __volatile__ ("movl %0,%%cr0": :"r" (x));

#define save_fpenv(x)    __asm__ __volatile__("fnsave %0" : "=m" (x))

#define restore_fpenv(x) __asm__ __volatile__("frstor %0" : "=m" (x));

static inline void rt_spin_lock_irq(spinlock_t *lock)          
{
	hard_cli(); 
	rt_spin_lock(lock);
}

static inline void rt_spin_unlock_irq(spinlock_t *lock)
{
	rt_spin_unlock(lock);
	hard_sti();
}

// Note that the spinlock calling convention below for irqsave/restore is 
// sligtly different from the one used in Linux. Done on purpose to get an 
// error if you use Linux spinlocks in real time applications as they do not
// guaranty any protection because of the soft irq disable. Be careful and 
// sure to call the other spinlocks the right way, as they are compatible 
// with Linux.

static inline unsigned int rt_spin_lock_irqsave(spinlock_t *lock)          
{
	unsigned int flags;
	hard_save_flags(flags);
	hard_cli(); 
	rt_spin_lock(lock);
	return flags;
}

static inline void rt_spin_unlock_irqrestore(unsigned int flags,
						spinlock_t *lock)
{
	rt_spin_unlock(lock);
	hard_restore_flags(flags);
}


#define RT_TIME_END 0x7FFFFFFFFFFFFFFFLL

typedef long long RTIME;

struct rt_times {  int linux_tick,
                       periodic_tick;
		 RTIME tick_time, 
		       linux_time,
		       intr_time; };

struct calibration_data { unsigned int cpu_freq;
				   int latency,
				       setup_time_8254CPU,
				       setup_time_8254;    };

extern struct rt_times rt_times;
extern struct calibration_data tuned;

extern RTIME rd_8254_ts(void);
extern void rt_setup_8254_tsc(void);

#undef rdtsc
#if CPU_FREQ == FREQ_8254
#define rdtsc() rd_8254_ts()
#else
#define rdtsc() rd_CPU_ts()
#endif

// get the time stamp clock (TSC) of a cpu.
static inline unsigned long long rd_CPU_ts(void)
{
	union {unsigned long long time; unsigned long time_lh[2]; } tsc;
	__asm__ __volatile__( "rdtsc"
			      : "=a" (tsc.time_lh[0]), "=d" (tsc.time_lh[1])
			      : 
			      :"ax", "dx");

	return tsc.time;
}

// returns (int)i = (int)i*(int)(mult)/(int)div.
static inline int imuldiv(int i, int mult, int div)
{
       __asm__("imull %%edx ;     idiv %%ecx; sal $1,%%edx  \n\t"
	       "cmpl %%edx,%%ecx; jge 1f;     incl %%eax;   1:"
               : "=a" (i)
       	       : "a" (i), "d" (mult), "c" (div)
               : "%eax", "%ecx", "%edx");

       return i;
}

// returns (long long)ll = (int)ll*(int)(mult)/(int)div.
static inline long long llimd(long long ll, int mult, int div)
{
	__asm__( "movl %%edx,%%ecx; mull %%esi;       movl %%eax,%%ebx;  \n\t"
	         "movl %%ecx,%%eax; movl %%edx,%%ecx; mull %%esi;        \n\t"
		 "addl %%ecx,%%eax; adcl $0,%%edx;    divl %%edi;        \n\t"
	         "movl %%eax,%%ecx; movl %%ebx,%%eax; divl %%edi;        \n\t"
		 "sal $1,%%edx;     cmpl %%edx,%%edi; movl %%ecx,%%edx;  \n\t"
		 "jge 1f;           addl $1,%%eax;    adcl $0,%%edx;     1:"
		 : "=A" (ll)
		 : "A" (ll), "S" (mult), "D" (div)
		 : "%eax", "%ebx", "%ecx", "%edx", "%esi", "%edi" );

	return ll;
}

// copied from Linux ffz (simply by taking ~ away).
static __inline__ unsigned long ffnz(unsigned long word)
{
	__asm__("bsfl %1, %0"
		: "=r" (word)
		: "r"  (word));

	return word;
}
